#include "Variable/VariableAcquisitionWorker.h"

#include "Variable/Variable.h"

#include <Data/IDataProvider.h>
#include <Data/SqpRange.h>
#include <unordered_map>
#include <utility>

#include <QMutex>
#include <QReadWriteLock>
#include <QThread>
#include <QtConcurrent/QtConcurrent>

Q_LOGGING_CATEGORY(LOG_VariableAcquisitionWorker, "VariableAcquisitionWorker")

namespace {

using AcquisitionId = QUuid;
using VariableId = QUuid;

struct Acquisition {
    std::shared_ptr<Variable> m_Variable{nullptr};
    VariableRequest m_Request{};
    AcquisitionId m_Id{QUuid::createUuid()};
};

} // namespace

struct VariableAcquisitionWorker::VariableAcquisitionWorkerPrivate {

    explicit VariableAcquisitionWorkerPrivate() : m_Lock{QReadWriteLock::Recursive} {}

    void lockRead() { m_Lock.lockForRead(); }
    void lockWrite() { m_Lock.lockForWrite(); }
    void unlock() { m_Lock.unlock(); }

    void eraseAcquisition(AcquisitionId id)
    {
        auto it = m_Acquisitions.find(id);
        if (it != m_Acquisitions.end()) {
            // Removes from index
            m_AcquisitionsIndex.erase(it->second.m_Variable);

            // Removes request
            m_Acquisitions.erase(it);
        }
    }

    std::map<AcquisitionId, Acquisition>::iterator insertAcquisition(Acquisition acquisition)
    {
        auto variable = acquisition.m_Variable;

        // Inserts acquisition
        auto result
            = m_Acquisitions.insert(std::make_pair(acquisition.m_Id, std::move(acquisition)));
        if (result.second) {
            // Inserts index
            m_AcquisitionsIndex[variable] = &result.first->second;
            return result.first;
        }
        else {
            return m_Acquisitions.end();
        }
    }

    QMutex m_WorkingMutex;
    QReadWriteLock m_Lock;

    std::map<AcquisitionId, Acquisition> m_Acquisitions;
    std::map<std::shared_ptr<Variable>, Acquisition *> m_AcquisitionsIndex;
};

VariableAcquisitionWorker::VariableAcquisitionWorker(QObject *parent)
        : QObject{parent}, impl{spimpl::make_unique_impl<VariableAcquisitionWorkerPrivate>()}
{
}

VariableAcquisitionWorker::~VariableAcquisitionWorker()
{
    qCInfo(LOG_VariableAcquisitionWorker())
        << tr("VariableAcquisitionWorker destruction") << QThread::currentThread();
    this->waitForFinish();
}

void VariableAcquisitionWorker::cancelVariableRequest(std::shared_ptr<Variable> variable)
{
    /// @todo ALX
}

void VariableAcquisitionWorker::pushVariableRequest(std::shared_ptr<Variable> variable,
                                                    VariableRequest request)
{
    impl->lockWrite();

    // Checks if there is a current request for variable
    auto oldAcquisitionIt = impl->m_AcquisitionsIndex.find(variable);
    if (oldAcquisitionIt != impl->m_AcquisitionsIndex.cend()) {
        auto &oldAcquisition = *oldAcquisitionIt->second;
        /// @todo ALX
        //        QtConcurrent::run(
        //            [ provider = request->m_Provider, acqIdentifier = request->m_AcqIdentifier ]()
        //            {
        //                provider->requestDataAborting(acqIdentifier);
        //            });

        impl->eraseAcquisition(oldAcquisition.m_Id);
    }

    // Sets request for variable
    Acquisition newAcquisition{variable, std::move(request)};
    auto newAcquisitionIt = impl->insertAcquisition(std::move(newAcquisition));
    if (newAcquisitionIt != impl->m_Acquisitions.end()) {
        impl->unlock();

        QMetaObject::invokeMethod(this, "executeAcquisition", Qt::QueuedConnection,
                                  Q_ARG(QUuid, newAcquisitionIt->first));
    }
    else {
        impl->unlock();
    }
}

void VariableAcquisitionWorker::abortProgressRequested(QUuid vIdentifier)
{
    // TODO
}

void VariableAcquisitionWorker::onVariableRetrieveDataInProgress(QUuid acqIdentifier,
                                                                 double progress)
{
    // TODO
}

void VariableAcquisitionWorker::onDataAcquired(QUuid acquisitionId,
                                               std::shared_ptr<IDataSeries> dataSeries,
                                               SqpRange range)
{
    impl->lockWrite();

    auto it = impl->m_Acquisitions.find(acquisitionId);
    if (it != impl->m_Acquisitions.cend()) {
        auto &acquisition = it->second;
        auto &request = acquisition.m_Request;

        // Store the result
        request.addResult(dataSeries);

        if (request.isFinished()) {
            emit dataProvided(acquisition.m_Variable, std::move(request));
            impl->eraseAcquisition(acquisitionId);
        }
    }
    impl->unlock();
}

void VariableAcquisitionWorker::onVariableAcquisitionCanceled(QUuid acqIdentifier)
{
}

void VariableAcquisitionWorker::initialize()
{
    qCDebug(LOG_VariableAcquisitionWorker())
        << tr("VariableAcquisitionWorker init") << QThread::currentThread();
    impl->m_WorkingMutex.lock();
    qCDebug(LOG_VariableAcquisitionWorker()) << tr("VariableAcquisitionWorker init END");
}

void VariableAcquisitionWorker::finalize()
{
    impl->m_WorkingMutex.unlock();
}

void VariableAcquisitionWorker::waitForFinish()
{
    QMutexLocker locker{&impl->m_WorkingMutex};
}

void VariableAcquisitionWorker::executeAcquisition(QUuid acquisitionId)
{
    impl->lockRead();
    auto it = impl->m_Acquisitions.find(acquisitionId);
    if (it != impl->m_Acquisitions.cend()) {
        auto &request = it->second.m_Request;
        impl->unlock();
        request.m_Provider->requestDataLoading(acquisitionId, request.m_ProviderParameters);
    }
    else {
        impl->unlock();
        // TODO log no acqIdentifier recognized
    }
}
